@@ -24,7 +24,7 @@
 
 use std::{
     cmp::max,
-    collections::HashMap,
+    collections::{BTreeSet, HashMap},
     fmt::{self, Display},
     ops::{Bound, RangeBounds},
 };
@@ -47,9 +47,10 @@ const DEFAULT_SPLIT_INTERSECTION_CHAR: char = ' ';
 /// Grid provides a set of methods for building a text-based table
 pub struct Grid {
     size: (usize, usize),
-    styles: HashMap<Entity, Style>,
     cells: Vec<Vec<String>>,
+    styles: HashMap<Entity, Style>,
     borders: Borders,
+    override_split_lines: HashMap<usize, String>,
 }
 
 impl Grid {
@@ -82,6 +83,7 @@ impl Grid {
             cells: vec![vec![String::new(); columns]; rows],
             styles,
             borders: Borders::new(rows, columns),
+            override_split_lines: HashMap::new(),
         }
     }
 
@@ -193,6 +195,10 @@ impl Grid {
         self.borders.clear()
     }
 
+    pub fn clear_overide_split_lines(&mut self) {
+        self.override_split_lines.clear();
+    }
+
     fn set_border(&mut self, frame: &EntityFrame, border: Border) {
         if let Some(top) = border.top {
             for column in frame.left_column..frame.right_column {
@@ -420,6 +426,10 @@ impl Grid {
         new_grid
     }
 
+    pub fn override_split_line(&mut self, row: usize, line: impl Into<String>) {
+        self.override_split_lines.insert(row, line.into());
+    }
+
     fn add_split_lines_for_border(&mut self, frame: &EntityFrame, border: &Border) {
         if border.left.is_some() && !self.is_vertical_present(frame.left_column) {
             self.add_vertical_split(frame.left_column)
@@ -819,12 +829,22 @@ impl std::fmt::Display for Grid {
         let mut styles = self.collect_styles(count_rows, count_columns);
 
         let row_heights = rows_height(&cells, &styles, count_rows, count_columns);
-        let widths = __columns_width(&mut cells, &mut styles, count_rows, count_columns);
+        let widths = columns_width(&mut cells, &mut styles, count_rows, count_columns);
 
         for row in 0..count_rows {
-            let top_border = self.get_split_line(row);
             let inner_border = self.get_inner_split_line(row);
-            let bottom_border = self.get_split_line(row + 1);
+            let top_border = if row == 0 {
+                Some((
+                    self.get_split_line(row),
+                    self.override_split_lines.get(&row),
+                ))
+            } else {
+                None
+            };
+            let bottom_border = Some((
+                self.get_split_line(row + 1),
+                self.override_split_lines.get(&(row + 1)),
+            ));
 
             build_row(
                 f,
@@ -832,8 +852,9 @@ impl std::fmt::Display for Grid {
                 &styles[row],
                 &widths[row],
                 row_heights[row],
-                row == 0,
-                (&top_border, &inner_border, &bottom_border),
+                inner_border,
+                top_border,
+                bottom_border,
             )?;
         }
 
@@ -841,17 +862,19 @@ impl std::fmt::Display for Grid {
     }
 }
 
+#[allow(clippy::too_many_arguments)]
 fn build_row(
     f: &mut std::fmt::Formatter<'_>,
     cell_contents: &[Vec<&str>],
     cell_styles: &[Style],
     cell_widths: &[usize],
     height: usize,
-    is_first_row: bool,
-    (top_borders, inner_borders, bottom_borders): (&[BorderLine], &[BorderLine], &[BorderLine]),
+    inner_border: Vec<BorderLine>,
+    top_border: Option<(Vec<BorderLine>, Option<&String>)>,
+    bottom_border: Option<(Vec<BorderLine>, Option<&String>)>,
 ) -> fmt::Result {
-    if is_first_row {
-        build_split_line(f, cell_widths, top_borders)?;
+    if let Some((top_border, override_border)) = top_border {
+        build_split_line(f, cell_widths, &top_border, override_border)?;
     }
 
     build_row_internals(
@@ -860,10 +883,13 @@ fn build_row(
         cell_styles,
         cell_widths,
         height,
-        inner_borders,
+        &inner_border,
     )?;
 
-    build_split_line(f, cell_widths, bottom_borders)?;
+    if let Some((bottom_border, override_border)) = bottom_border {
+        build_split_line(f, cell_widths, &bottom_border, override_border)?;
+    }
+
     Ok(())
 }
 
@@ -955,11 +981,11 @@ fn line(
     Ok(())
 }
 
-fn build_line<F: Fn(&mut std::fmt::Formatter<'_>, usize) -> fmt::Result>(
+fn build_line<F: FnMut(&mut std::fmt::Formatter<'_>, usize) -> fmt::Result>(
     f: &mut std::fmt::Formatter<'_>,
     length: usize,
     borders: &[BorderLine],
-    writer: F,
+    mut writer: F,
 ) -> fmt::Result {
     for (i, border) in borders.iter().enumerate().take(length) {
         write_option(f, border.connector1)?;
@@ -978,15 +1004,69 @@ fn build_split_line(
     f: &mut std::fmt::Formatter<'_>,
     widths: &[usize],
     borders: &[BorderLine],
+    override_str: Option<&String>,
 ) -> fmt::Result {
     let theres_no_border = borders.iter().all(|l| l.main.is_none());
     if theres_no_border || borders.is_empty() {
         return Ok(());
     }
 
-    build_line(f, widths.len(), borders, |f, i| {
-        write_option(f, borders[i].main.map(|m| m.to_string().repeat(widths[i])))
-    })
+    let mut override_str = override_str.map(|s| s.to_owned());
+    for (i, border) in borders.iter().enumerate().take(widths.len()) {
+        if let Some(left_connector) = border.connector1 {
+            let connector = override_str
+                .as_mut()
+                .and_then(|s| {
+                    s.chars().next().map(|c| {
+                        let _ = s.drain(..c.len_utf8());
+                        c
+                    })
+                })
+                .unwrap_or(left_connector);
+            write!(f, "{}", connector)?
+        }
+
+        if let Some(main) = border.main {
+            let mut width = widths[i];
+            if let Some(s) = override_str.as_mut() {
+                while !s.is_empty() && width > 0 {
+                    match s.chars().next() {
+                        Some(c) => {
+                            write!(f, "{}", c)?;
+                            width -= 1;
+                            let _ = s.drain(..c.len_utf8());
+                        }
+                        None => break,
+                    }
+                }
+            }
+
+            while width > 0 {
+                write!(f, "{}", main)?;
+                width -= 1;
+            }
+        }
+
+        let is_last_cell = i + 1 == widths.len();
+        if is_last_cell {
+            if let Some(right_connector) = border.connector2 {
+                let connector = override_str
+                    .as_mut()
+                    .and_then(|s| {
+                        s.chars().next().map(|c| {
+                            let _ = s.drain(..c.len_utf8());
+                            c
+                        })
+                    })
+                    .unwrap_or(right_connector);
+                write!(f, "{}", connector)?
+            }
+        }
+    }
+
+    writeln!(f)?;
+
+    Ok(())
 }
 
 fn write_option<D: Display>(f: &mut std::fmt::Formatter<'_>, text: Option<D>) -> fmt::Result {
@@ -1015,13 +1095,13 @@ fn real_string_width(text: &str) -> usize {
         .unwrap_or(0)
 }
 
-fn __columns_width(
+fn columns_width(
     cells: &mut [Vec<Vec<&str>>],
     styles: &mut [Vec<Style>],
     count_rows: usize,
     count_columns: usize,
 ) -> Vec<Vec<usize>> {
-    let mut widths = vec![vec! {0; count_columns}; count_rows];
+    let mut widths = vec![vec![0; count_columns]; count_rows];
     (0..count_rows).for_each(|row| {
         (0..count_columns).for_each(|column| {
             let cell = &cells[row][column];
@@ -1030,145 +1110,170 @@ fn __columns_width(
                 widths[row][column] = cell_width(cell, style);
             } else {
                 widths[row][column] = 0;
+                styles[row][column].span = 0;
             }
         });
     });
 
-    // check if we don't need to check all spans as it a heavy load function.
-    // it suppose to save us time and resources
-    let spans_was_used = styles
-        .iter()
-        .any(|row_styles| row_styles.iter().any(|style| style.span > 1));
-    if spans_was_used {
-        (1..count_columns + 1).for_each(|span| {
-            __adjust_width(&mut widths, cells, styles, count_rows, count_columns, span);
-        });
-    } else {
-        __adjust_width(&mut widths, cells, styles, count_rows, count_columns, 1);
-    }
+    // it's crusial to preserve order in iterations
+    // so we use BTreeSet
+    let mut spans = BTreeSet::new();
+    styles.iter().for_each(|row_styles| {
+        row_styles.iter().for_each(|style| {
+            spans.insert(style.span);
+        })
+    });
+    spans.into_iter().filter(|&span| span > 0).for_each(|span| {
+        adjust_width(&mut widths, styles, count_rows, count_columns, span);
+    });
 
     // remove not visible cells to print everything correctly
     (0..count_rows).for_each(|row| {
         let mut n_removed = 0;
-        (0..count_columns)
-            .filter(|&column| !is_cell_visible(&styles[row], column))
-            .collect::<Vec<_>>() // it's here becouse of borrow rules...
-            .into_iter()
-            .for_each(|column| {
-                widths[row].remove(column - n_removed);
-                cells[row].remove(column - n_removed);
-                styles[row].remove(column - n_removed);
+        (0..count_columns).for_each(|column| {
+            let column = column - n_removed;
+            if styles[row][column].span == 0 {
+                widths[row].remove(column);
+                cells[row].remove(column);
+                styles[row].remove(column);
                 n_removed += 1;
-            });
+            }
+        });
     });
 
     widths
 }
 
-fn __adjust_width(
+fn adjust_width(
     widths: &mut [Vec<usize>],
-    cells: &[Vec<Vec<&str>>],
     styles: &[Vec<Style>],
     count_rows: usize,
     count_columns: usize,
     span: usize,
 ) {
-    (0..count_rows).for_each(|row| {
-        (0..count_columns)
-            .filter(|&column| styles[row][column].span == span)
-            .filter(|&column| is_cell_visible(&styles[row], column))
-            .for_each(|column| {
-                let cell = &cells[row][column];
-                let style = &styles[row][column];
-                let cell_width = cell_width(cell, style);
-                // calc other's width
+    for column in 0..count_columns {
+        let start = column;
+        let end = column + span;
+        if end > count_columns {
+            break;
+        }
 
-                let others_width = (0..count_rows)
-                    .filter(|&r| r != row)
-                    .filter(|&r| {
-                        styles[r][column..column + span]
-                            .iter()
-                            .map(|style| style.span)
-                            .sum::<usize>()
-                            <= span
-                    })
-                    .map(|r| row_width(&styles[r], &widths[r][column..column + span]))
-                    .max()
-                    .unwrap_or(0);
+        adjust_range_width(widths, styles, count_rows, start, end);
+    }
+}
 
-                if cell_width > others_width {
-                    widths[row][column] = cell_width;
+fn adjust_range_width(
+    widths: &mut [Vec<usize>],
+    styles: &[Vec<Style>],
+    count_rows: usize,
+    start_column: usize,
+    end_column: usize,
+) {
+    if count_rows == 0 {
+        return;
+    }
+    let span = end_column - start_column;
 
-                    (0..count_rows)
-                        .filter(|&r| r != row)
-                        .filter(|&r| {
-                            styles[r][column..column + span]
-                                .iter()
-                                .map(|style| style.span)
-                                .sum::<usize>()
-                                <= span
-                        }) // not sure if it's correct
-                        .for_each(|r| {
-                            inc_width_to_cells(
-                                &mut widths[r],
-                                &styles[r],
-                                column,
-                                column + span,
-                                cell_width,
-                            );
-                        });
-                } else {
-                    inc_width_to_cells(
-                        &mut widths[row],
-                        &styles[row],
-                        column,
-                        column + 1,
-                        others_width,
-                    );
-                }
-            });
-    });
+    let (row_with_max_width, max_width) = (0..count_rows)
+        .filter(|&row| is_row_consistent(&styles[row][start_column..end_column]))
+        .filter(|&row| {
+            let row_spans = (start_column..end_column)
+                .filter(|&column| is_cell_visible(&styles[row], column))
+                .map(|col| styles[row][col].span)
+                .sum::<usize>();
+
+            row_spans <= span
+        })
+        .map(|row| {
+            let width = row_width(&styles[row], &widths[row], start_column, end_column);
+            (row, width)
+        })
+        .max_by_key(|&(_, width)| width)
+        .unwrap_or_default();
+
+    // might happen when we filtered every cell
+    if max_width == 0 {
+        return;
+    }
+
+    (0..count_rows)
+        .filter(|&row| row != row_with_max_width)
+        .filter(|&row| is_row_consistent(&styles[row][start_column..end_column]))
+        .filter(|&row| {
+            let row_spans = (start_column..end_column)
+                .filter(|&column| is_cell_visible(&styles[row], column))
+                .map(|col| styles[row][col].span)
+                .sum::<usize>();
+
+            row_spans <= span
+        })
+        .for_each(|row| {
+            inc_width_to_cells(
+                &mut widths[row],
+                &styles[row],
+                start_column,
+                end_column,
+                max_width,
+            );
+        });
+}
+
+fn is_row_consistent(styles: &[Style]) -> bool {
+    if styles.is_empty() {
+        return true;
+    }
+
+    if styles[0].span == 0 {
+        return false;
+    }
+
+    styles
+        .iter()
+        .zip(styles.len()..)
+        .all(|(style, max_span)| style.span <= max_span)
 }
 
 fn is_cell_visible(row_styles: &[Style], column: usize) -> bool {
-    !row_styles[..column]
-        .iter()
-        .zip(column..)
-        .any(|(style, span)| style.span > span)
+    row_styles[column].span != 0
+        && !row_styles[..column]
+            .iter()
+            .zip(column..)
+            .any(|(style, span)| style.span > span)
 }
 
-// relyes on fix_spans
-fn row_width(row_styles: &[Style], widths: &[usize]) -> usize {
-    let w = (0..widths.len())
-        .filter(|&i| is_cell_visible(row_styles, i))
+fn row_width(styles: &[Style], widths: &[usize], column_start: usize, column_end: usize) -> usize {
+    let width = (column_start..column_end)
+        .filter(|&i| is_cell_visible(styles, i))
         .map(|i| widths[i])
         .sum::<usize>();
 
-    if w == 0 {
+    if width == 0 {
         return 0;
     }
 
-    w + (0..widths.len())
-        .filter(|&i| is_cell_visible(row_styles, i))
+    // fixme: Is this style dependent?
+    let border_count = (column_start..column_end)
+        .filter(|&i| is_cell_visible(styles, i))
         .count()
-        - 1
+        - 1;
+
+    width + border_count
 }
 
-// relyes on fix_spans
 fn inc_width_to_cells(
     widths: &mut [usize],
-    row_styles: &[Style],
+    styles: &[Style],
     start_range: usize,
     end_range: usize,
     width: usize,
 ) {
-    let a = row_width(row_styles, &widths[start_range..end_range]);
+    let a = row_width(styles, widths, start_range, end_range);
     let diff = width - a;
 
     (0..diff)
         .zip(
             (start_range..end_range)
-                .filter(|&i| is_cell_visible(row_styles, i))
+                .filter(|&i| is_cell_visible(styles, i))
                 .cycle(),
         )
         .for_each(|(_, i)| widths[i] += 1);
